# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Define enum to describe monotonicity of expressions."""
from enum import Enum


class Monotonicity(Enum):
    """Monotonicity information.

    Monotonicity can be:

     * Nondecreasing: f(x) >= f(y) for all x > y
     * Nonincreasing: f(x) <= f(y) for all x > y
     * Constant: f(x) = f(y) for all x > y
     * Unknown
    """
    Nondecreasing = 0
    Nonincreasing = 1
    Constant = 2
    Unknown = 3

    def is_nondecreasing(self):
        """Predicate to check if monotonicity is nondecreasing."""
        return self == self.Nondecreasing or self == self.Constant

    def is_nonincreasing(self):
        """Predicate to check if monotonicity is nonincreasing."""
        return self == self.Nonincreasing or self == self.Constant

    def is_constant(self):
        """Predicate to check if monotonicity is constant."""
        return self == self.Constant

    def is_unknown(self):
        """Predicate to check if monotonicity is unknown."""
        return self == self.Unknown

    def negate(self):
        """Return monotonicity of the negated expression.

        * If the function is constant, the negation will be constant
        * If the function is unknown, the negation will be unknown
        * If the function is nondecreasing, the negation will be nonincreasing
        * If the function is nonincreasing, the negation will be nondecreasing
        """
        if self.is_constant() or self.is_unknown():
            return self
        if self.is_nondecreasing():
            return Monotonicity.Nonincreasing
        return Monotonicity.Nondecreasing



# pylint: skip-file
import pytest
from hypothesis import given, assume, reproduce_failure, settings, HealthCheck
import hypothesis.strategies as st
import itertools
from pyomo.core.kernel.component_map import ComponentMap
import pyomo.environ as pe
from tests.strategies import coefficients, reals, expressions
from tests.conftest import PlaceholderExpression as PE
from suspect.expression import ExpressionType as ET, UnaryFunctionType as UFT
from suspect.monotonicity.visitor import MonotonicityPropagationVisitor
from suspect.monotonicity.monotonicity import Monotonicity as M
from suspect.pyomo.expressions import (
    NumericConstant,
    ProductExpression,
    PowExpression,
    SumExpression,
    NegationExpression,
    DivisionExpression,
    Constraint,
    Objective,
    Sense,
)
from suspect.interval import Interval as I
from suspect.monotonicity.rules import *
from suspect.math import almosteq, pi


@pytest.fixture
def visitor():
    return MonotonicityPropagationVisitor()


@st.composite
def nondecreasing_terms(draw):
    coef = draw(coefficients())
    if coef > 0:
        return coef, M.Nondecreasing
    return coef, M.Nonincreasing


@st.composite
def nonincreasing_terms(draw):
    coef = draw(coefficients())
    if coef < 0:
        return coef, M.Nondecreasing
    return coef, M.Nonincreasing


@st.composite
def constant_terms(draw):
    coef = draw(coefficients())
    return coef, M.Constant


@st.composite
def nonnegative_cos_bounds(draw):
    start = draw(st.floats(
        min_value=1.6*pi,
        max_value=1.9*pi,
        allow_nan=False,
        allow_infinity=False,
    ))
    end = draw(st.floats(
        min_value=0,
        max_value=1.9*pi-start,
        allow_nan=False,
        allow_infinity=False,
    ))
    mul = draw(st.integers(min_value=-100, max_value=100)) * 2 * pi
    b = I(start + mul, start + end + mul)
    assume(b.size() > 0)
    return b


@st.composite
def nonpositive_cos_bounds(draw):
    start = draw(st.floats(
        min_value=0.6*pi,
        max_value=1.5*pi,
        allow_nan=False,
        allow_infinity=False,
    ))
    end = draw(st.floats(
        min_value=0,
        max_value=1.5*pi-start,
        allow_nan=False,
        allow_infinity=False,
    ))
    mul = draw(st.integers(min_value=-10, max_value=10)) * 2 * pi
    b = I(start + mul, start + end + mul)
    assume(b.size() > 0)
    return b


@st.composite
def nonnegative_sin_bounds(draw):
    start = draw(st.floats(
        min_value=0.1*pi,
        max_value=0.9*pi,
        allow_nan=False,
        allow_infinity=False,
    ))
    end = draw(st.floats(
        min_value=0,
        max_value=0.9*pi-start,
        allow_nan=False,
        allow_infinity=False,
    ))
    mul = draw(st.integers(min_value=-100, max_value=100)) * 2 * pi
    b = I(start + mul, start + end + mul)
    assume(b.size() > 0)
    return b


@st.composite
def nonpositive_sin_bounds(draw):
    start = draw(st.floats(
        min_value=1.1*pi,
        max_value=1.9*pi,
        allow_nan=False,
        allow_infinity=False,
    ))
    end = draw(st.floats(
        min_value=0,
        max_value=1.9*pi-start,
        allow_nan=False,
        allow_infinity=False,
    ))
    mul = draw(st.integers(min_value=-10, max_value=10)) * 2 * pi
    b = I(start + mul, start + end + mul)
    assume(b.size() > 0)
    return b


def test_variable_is_nondecreasing():
    rule = VariableRule()
    result = rule.apply(PE(ET.Variable), None, None)
    assert result.is_nondecreasing() and (not result.is_constant())


def test_constant_is_constant():
    rule = ConstantRule()
    result = rule.apply(PE(ET.Constant), None, None)
    assert result.is_constant()


@pytest.mark.parametrize('expr_mono,expected_mono', [
    (M.Nondecreasing, M.Nondecreasing),
    (M.Nonincreasing, M.Nonincreasing),
    (M.Constant, M.Constant),
    (M.Unknown, M.Unknown),
])
@given(child=expressions())
def test_lte_constraint(visitor, child, expr_mono, expected_mono):
    mono = ComponentMap()
    mono[child] = expr_mono
    expr = Constraint('cons', None, 0.0, children=[child])
    matched, result = visitor.visit_expression(expr, mono, None)
    assert matched
    assert result == expected_mono


@pytest.mark.parametrize('expr_mono,expected_mono', [
    (M.Nondecreasing, M.Nonincreasing),
    (M.Nonincreasing, M.Nondecreasing),
    (M.Constant, M.Constant),
    (M.Unknown, M.Unknown),
])
@given(child=expressions())
def test_gte_constraint(visitor, child, expr_mono, expected_mono):
    mono = ComponentMap()
    mono[child] = expr_mono
    expr = Constraint('cons', 0.0, None, children=[child])
    matched, result = visitor.visit_expression(expr, mono, None)
    assert matched
    assert result == expected_mono


@pytest.mark.parametrize('expr_mono,expected_mono', [
    (M.Nondecreasing, M.Unknown),
    (M.Nonincreasing, M.Unknown),
    (M.Constant, M.Constant),
    (M.Unknown, M.Unknown),
])
@given(child=expressions())
def test_eq_constraint(visitor, child, expr_mono, expected_mono):
    mono = ComponentMap()
    mono[child] = expr_mono
    expr = Constraint('cons', 0.0, 0.0, children=[child])
    matched, result = visitor.visit_expression(expr, mono, None)
    assert matched
    assert result == expected_mono


@pytest.mark.parametrize('expr_mono,expected_mono', [
    (M.Nondecreasing, M.Nondecreasing),
    (M.Nonincreasing, M.Nonincreasing),
    (M.Constant, M.Constant),
    (M.Unknown, M.Unknown),
])
@given(child=expressions())
def test_min_constraint(visitor, child, expr_mono, expected_mono):
    mono = ComponentMap()
    mono[child] = expr_mono
    expr = Objective('obj', children=[child])
    matched, result = visitor.visit_expression(expr, mono, None)
    assert matched
    assert result == expected_mono


@pytest.mark.parametrize('expr_mono,expected_mono', [
    (M.Nondecreasing, M.Nonincreasing),
    (M.Nonincreasing, M.Nondecreasing),
    (M.Constant, M.Constant),
    (M.Unknown, M.Unknown),
])
@given(child=expressions())
def test_max_constraint(visitor, child, expr_mono, expected_mono):
    mono = ComponentMap()
    mono[child] = expr_mono
    expr = Objective('obj', children=[child], sense=Sense.MAXIMIZE)
    matched, result = visitor.visit_expression(expr, mono, None)
    assert matched
    assert result == expected_mono


@pytest.mark.parametrize('mono_f,mono_g,bound_f,bound_g,expected', [
    (M.Nondecreasing, M.Constant, I(None, None), I(0, None), M.Nondecreasing),
    (M.Nonincreasing, M.Constant, I(None, None), I(None, 0), M.Nondecreasing),

    (M.Nondecreasing, M.Constant, I(None, None), I(None, 0), M.Nonincreasing),
    (M.Nonincreasing, M.Constant, I(None, None), I(0, None), M.Nonincreasing),

    (M.Constant, M.Nondecreasing, I(0, None), I(None, None), M.Nondecreasing),
    (M.Constant, M.Nonincreasing, I(None, 0), I(None, None), M.Nondecreasing),

    (M.Constant, M.Nondecreasing, I(None, 0), I(None, None), M.Nonincreasing),
    (M.Constant, M.Nonincreasing, I(0, None), I(None, None), M.Nonincreasing),

    (M.Constant, M.Constant, I(None, None), I(None, None), M.Constant),

    (M.Nondecreasing, M.Nondecreasing, I(0, None), I(0, None), M.Nondecreasing),
    (M.Nondecreasing, M.Nonincreasing, I(None, 0), I(0, None), M.Nondecreasing),
    (M.Nonincreasing, M.Nondecreasing, I(0, None), I(None, 0), M.Nondecreasing),
    (M.Nonincreasing, M.Nonincreasing, I(None, 0), I(None, 0), M.Nondecreasing),

    (M.Nonincreasing, M.Nonincreasing, I(0, None), I(0, None), M.Nonincreasing),
    (M.Nonincreasing, M.Nondecreasing, I(None, 0), I(0, None), M.Nonincreasing),
    (M.Nondecreasing, M.Nonincreasing, I(0, None), I(None, 0), M.Nonincreasing),
    (M.Nondecreasing, M.Nondecreasing, I(None, 0), I(None, 0), M.Nonincreasing),
])
@given(f=expressions(), g=expressions())
def test_product(visitor, f, g, mono_f, mono_g, bound_f, bound_g, expected):
    bounds = ComponentMap()
    bounds[f] = bound_f
    bounds[g] = bound_g
    mono = ComponentMap()
    mono[f] = mono_f
    mono[g] = mono_g

    expr = f * g
    assume(isinstance(expr, ProductExpression))
    matched, result = visitor.visit_expression(expr, mono, bounds)
    assert matched
    assert result == expected


@pytest.mark.skip('Pyomo has no division expression')
@pytest.mark.parametrize('mono_f,mono_g,bound_f,bound_g,expected', [
    (M.Constant, M.Constant, I(1.0, 1.0), I(2.0, 2.0), M.Constant),
    (M.Constant, M.Constant, I(1.0, 1.0), I(0.0, 0.0), M.Unknown),

    (M.Nondecreasing, M.Constant, I(None, None), I(10.0, 10.0), M.Nondecreasing),
    (M.Nonincreasing, M.Constant, I(None, None), I(-1.0, -1.0), M.Nondecreasing),

    (M.Nondecreasing, M.Constant, I(None, None), I(-1.0, -1.0), M.Nonincreasing),
    (M.Nonincreasing, M.Constant, I(None, None), I(10.0, 10.0), M.Nonincreasing),

    (M.Nondecreasing, M.Nonincreasing, I(0, None), I(0, None), M.Nondecreasing),
    (M.Nondecreasing, M.Nondecreasing, I(None, 0), I(0, None), M.Nondecreasing),
    (M.Nonincreasing, M.Nonincreasing, I(0, None), I(None, 0), M.Nondecreasing),
    (M.Nonincreasing, M.Nondecreasing, I(None, 0), I(None, 0), M.Nondecreasing),

    (M.Nonincreasing, M.Nondecreasing, I(0, None), I(0, None), M.Nonincreasing),
    (M.Nonincreasing, M.Nonincreasing, I(None, 0), I(0, None), M.Nonincreasing),
    (M.Nondecreasing, M.Nondecreasing, I(0, None), I(None, 0), M.Nonincreasing),
    (M.Nondecreasing, M.Nonincreasing, I(None, 0), I(None, 0), M.Nonincreasing),
])
@given(f=expressions(), g=expressions())
def test_division(visitor, f, g, mono_f, mono_g, bound_f, bound_g, expected):
    bounds = ComponentMap()
    bounds[f] = bound_f
    bounds[g] = bound_g
    mono = ComponentMap()
    mono[f] = mono_f
    mono[g] = mono_g

    expr = f / g
    # assume(isinstance(expr, DivisionExpression))
    matched, result = visitor.visit_expression(expr, mono, bounds)
    assert matched
    assert result == expected


@pytest.mark.parametrize('mono_g,bound_g,expected', [
    (M.Constant, I(2.0, 2.0), M.Constant),
    (M.Constant, I(0.0, 0.0), M.Unknown),
])
@given(g=expressions())
def test_division(visitor, g, mono_g, bound_g, expected):
    num = NumericConstant(1.0)
    bounds = ComponentMap()
    bounds[num] = I(1.0, 1.0)
    bounds[g] = bound_g
    mono = ComponentMap()
    mono[num] = M.Constant
    mono[g] = mono_g

    print(mono)
    print(mono[num])
    expr = DivisionExpression([num, g])
    matched, result = visitor.visit_expression(expr, mono, bounds)
    assert matched
    assert result == expected


@pytest.mark.skip('Pyomo has no linear expressions')
class TestLinear(object):
    def _result_with_terms(self, terms):
        rule = LinearRule()
        monotonicity = {}
        coefficients = [c for c, _ in terms]
        children = [PE(ET.Variable) for _ in terms]
        monos = [m for _, m in terms]
        for child, mono in zip(children, monos):
            monotonicity[child] = mono
        ctx = MonotonicityContext(monotonicity, {})
        result = rule.checked_apply(
            PE(ET.Linear, children=children, coefficients=coefficients),
            ctx,
        )
        return result

    @given(
        st.lists(
            st.one_of(nondecreasing_terms(), constant_terms()),
            min_size=1)
    )
    def test_nondecreasing(self, terms):
        mono = self._result_with_terms(terms)
        assert mono.is_nondecreasing()

    @given(
        st.lists(
            st.one_of(nonincreasing_terms(), constant_terms()),
            min_size=1)
    )
    def test_nonincreasing(self, terms):
        mono = self._result_with_terms(terms)
        assert mono.is_nonincreasing()

    @given(st.lists(constant_terms(), min_size=1))
    def test_constant(self, terms):
        mono = self._result_with_terms(terms)
        assert mono.is_constant()

    @given(
        st.lists(nondecreasing_terms(), min_size=1),
        st.lists(nonincreasing_terms(), min_size=1),
    )
    def test_unknown(self, a, b):
        terms = a + b
        for c, _ in terms:
            assume(c != 0.0)
        mono = self._result_with_terms(terms)
        assert mono.is_unknown()



class TestSum(object):
    def _result_with_terms(self, visitor, expr_with_monos):
        children = [c for c, _ in expr_with_monos]
        monos = [m for _, m in expr_with_monos]
        monotonicity = ComponentMap()
        for child, mono in expr_with_monos:
            monotonicity[child] = mono
        expr = SumExpression(children)
        matched, result = visitor.visit_expression(expr, monotonicity, None)
        assert matched
        return result

    @given(
        st.lists(
            st.tuples(expressions(), st.just(M.Nondecreasing)),
            min_size=1,
        ),
        st.lists(
            st.tuples(
                expressions(),
                st.one_of(st.just(M.Nondecreasing), st.just(M.Constant))
            )
        ),
    )
    def test_nondecreasing(self, visitor, a, b):
        mono = self._result_with_terms(visitor, a + b)
        assert mono.is_nondecreasing()

    @given(
        st.lists(
            st.tuples(expressions(), st.just(M.Nonincreasing)),
            min_size=1
        ),
        st.lists(
            st.tuples(
                expressions(),
                st.one_of(st.just(M.Nonincreasing), st.just(M.Constant))
            )
        ),
    )
    def test_nonincreasing(self, visitor, a, b):
        mono = self._result_with_terms(visitor, a + b)
        assert mono.is_nonincreasing()

    @given(
        st.lists(
            st.tuples(
                expressions(),
                st.just(M.Constant)
            ),
            min_size=1
        ),
    )
    def test_constant(self, visitor, a):
        mono = self._result_with_terms(visitor, a)
        assert mono.is_constant()

    @given(
        st.lists(
            st.tuples(
                expressions(),
                st.just(M.Nondecreasing)
            ),
            min_size=1
        ),
        st.lists(
            st.tuples(
                expressions(),
                st.just(M.Nonincreasing),
            ),
            min_size=1
        ),
    )
    def test_unknown(self, visitor, a, b):
        mono = self._result_with_terms(visitor, a + b)
        assert mono.is_unknown()


@pytest.mark.parametrize('mono_g,bounds_g,expected', [
    (M.Nondecreasing, I(0, None), M.Nondecreasing),
    (M.Nonincreasing, I(None, 0), M.Nondecreasing),
    (M.Nonincreasing, I(0, None), M.Nonincreasing),
    (M.Nondecreasing, I(None, 0), M.Nonincreasing),
    (M.Constant, I(None, None), M.Constant),
    (M.Nondecreasing, I(None, None), M.Unknown),
    (M.Nonincreasing, I(None, None), M.Unknown),
])
@given(g=expressions())
def test_abs(visitor, g, mono_g, bounds_g, expected):
    bounds = ComponentMap()
    bounds[g] = bounds_g
    mono = ComponentMap()
    mono[g] = mono_g
    expr = abs(g)
    matched, result = visitor.visit_expression(expr, mono, bounds)
    assert matched
    assert result == expected


@pytest.mark.parametrize('func_name', [
    'sqrt', 'exp', 'log', 'tan', 'asin', 'atan',
])
@pytest.mark.parametrize('mono_g,bounds_g', itertools.product(
    [M.Nondecreasing, M.Nonincreasing, M.Constant, M.Unknown],
    [I(None, None), I(0, None), I(None, 0)],
))
@given(g=expressions())
def test_nondecreasing_function(visitor, g, func_name, mono_g, bounds_g):
    mono = ComponentMap()
    mono[g] = mono_g
    bounds = ComponentMap()
    bounds[g]= bounds_g
    func = getattr(pe, func_name)
    expr = func(g)
    matched, result = visitor.visit_expression(expr, mono, bounds)
    assert matched
    assert result == mono_g


@pytest.mark.parametrize('func_name', ['acos'])
@pytest.mark.parametrize('mono_g,bounds_g', itertools.product(
    [M.Nondecreasing, M.Nonincreasing, M.Constant, M.Unknown],
    [I(None, None), I(0, None), I(None, 0)],
))
@given(g=expressions())
def test_nonincreasing_function(visitor, g, func_name, mono_g, bounds_g):
    mono = ComponentMap()
    mono[g] = mono_g
    bounds = ComponentMap()
    bounds[g] = bounds_g
    func = getattr(pe, func_name)
    expr = func(g)
    matched, result = visitor.visit_expression(expr, mono, bounds)
    assert matched
    assert result == mono_g.negate()


@pytest.mark.parametrize('mono_g,bounds_g', itertools.product(
    [M.Nondecreasing, M.Nonincreasing, M.Constant, M.Unknown],
    [I(None, None), I(0, None), I(None, 0)],
))
@given(g=expressions())
def test_negation(visitor, g, mono_g, bounds_g):
    mono = ComponentMap()
    mono[g] = mono_g
    bounds = ComponentMap()
    bounds[g] = bounds_g
    expr = -g
    assume(isinstance(expr, NegationExpression))
    matched, result = visitor.visit_expression(expr, mono, bounds)
    assert matched
    assert result == mono_g.negate()


class TestPowConstantBase(object):
    def _result_with_base_expo(self, visitor, base, expo, mono_expo, bounds_expo):
        rule = PowerRule()
        mono = ComponentMap()
        mono[base] = M.Constant
        mono[expo] = mono_expo
        bounds = ComponentMap()
        bounds[base] = I(base, base)
        bounds[expo] = bounds_expo
        expr = base ** expo
        assume(isinstance(expr, PowExpression))
        matched, result = visitor.visit_expression(expr, mono, bounds)
        assert matched
        return result

    @pytest.mark.parametrize(
        'mono_expo,bounds_expo',
        itertools.product(
            [M.Nondecreasing, M.Nonincreasing, M.Unknown],
            [I(None, 0), I(0, None), I(None, None)]
        )
    )
    @given(
        base=reals(max_value=-0.01, allow_infinity=False),
        expo=expressions(),
    )
    def test_negative_base(self, visitor, base, expo, mono_expo, bounds_expo):
        mono = self._result_with_base_expo(visitor, base, expo, mono_expo, bounds_expo)
        assert mono == M.Unknown

    @pytest.mark.parametrize('mono_expo,bounds_expo,expected', [
        (M.Nondecreasing, I(None, 0), M.Nondecreasing),
        (M.Nondecreasing, I(0, None), M.Unknown),
        (M.Nonincreasing, I(0, None), M.Nondecreasing),
        (M.Nonincreasing, I(None, 0), M.Unknown),
    ])
    @given(
        base=reals(min_value=0.01, max_value=0.999),
        expo=expressions(),
    )
    def test_base_between_0_and_1(self, visitor, base, expo, mono_expo, bounds_expo, expected):
        mono = self._result_with_base_expo(visitor, base, expo, mono_expo, bounds_expo)
        assert mono == expected

    @pytest.mark.parametrize('mono_expo,bounds_expo,expected', [
        (M.Nondecreasing, I(0, None), M.Nondecreasing),
        (M.Nondecreasing, I(None, 0), M.Unknown),
        (M.Nonincreasing, I(None, 0), M.Nondecreasing),
        (M.Nonincreasing, I(0, None), M.Unknown),
    ])
    @given(
        base=reals(min_value=1.01, allow_infinity=False),
        expo=expressions(),
    )
    def test_base_gt_1(self, visitor, base, expo, mono_expo, bounds_expo, expected):
        mono = self._result_with_base_expo(visitor, base, expo, mono_expo, bounds_expo)
        assert mono == expected


class TestPowConstantExponent(object):
    def _result_with_base_expo(self, visitor, base, mono_base, bounds_base, expo):
        mono = ComponentMap()
        mono[base] = mono_base
        mono[expo] = M.Constant
        bounds = ComponentMap()
        bounds[base] = bounds_base
        bounds[expo] = I(expo, expo)
        expr = PowExpression([base, expo])
        matched, result = visitor.visit_expression(expr, mono, bounds)
        assert matched
        return result

    @pytest.mark.parametrize(
        'mono_base,bounds_base',
        itertools.product([M.Nonincreasing, M.Nondecreasing],
                          [I(None, 0), I(0, None), I(None, None)])
    )
    @given(base=expressions())
    def test_exponent_equals_1(self, visitor, base, mono_base, bounds_base):
        mono = self._result_with_base_expo(visitor, base, mono_base, bounds_base, 1.0)
        assert mono == mono_base

    @pytest.mark.parametrize(
        'mono_base,bounds_base',
        itertools.product([M.Nonincreasing, M.Nondecreasing],
                          [I(None, 0), I(0, None), I(None, None)])
    )
    @given(base=expressions())
    def test_exponent_equals_0(self, visitor, base, mono_base, bounds_base):
        mono = self._result_with_base_expo(visitor, base, mono_base, bounds_base, 0.0)
        assert mono == M.Constant

    @pytest.mark.parametrize('mono_base,bounds_base,expected', [
        (M.Nondecreasing, I(0, None), M.Nondecreasing),
        (M.Nonincreasing, I(None, 0), M.Nondecreasing),

        (M.Nondecreasing, I(None, 0), M.Nonincreasing),
        (M.Nonincreasing, I(0, None), M.Nonincreasing),
    ])
    @given(
        base=expressions(),
        expo=st.integers(min_value=1, max_value=1000),
    )
    def test_positive_even_integer(self, visitor, base, expo, mono_base, bounds_base, expected):
        mono = self._result_with_base_expo(visitor, base, mono_base, bounds_base, 2*expo)
        assert mono == expected

    @pytest.mark.parametrize('mono_base,bounds_base,expected', [
        (M.Nondecreasing, I(0, None), M.Nonincreasing),
        (M.Nonincreasing, I(None, 0), M.Nonincreasing),

        (M.Nondecreasing, I(None, 0), M.Nondecreasing),
        (M.Nonincreasing, I(0, None), M.Nondecreasing),
    ])
    @given(
        base=expressions(),
        expo=st.integers(min_value=1, max_value=1000)
    )
    def test_negative_even_integer(self, visitor, base, expo, mono_base, bounds_base, expected):
        mono = self._result_with_base_expo(visitor, base, mono_base, bounds_base, -2*expo)
        assert mono == expected

    @pytest.mark.parametrize('mono_base,expected', [
        (M.Nondecreasing, M.Nondecreasing),
        (M.Nonincreasing, M.Nonincreasing),
    ])
    @given(
        base=expressions(),
        expo=st.integers(min_value=1, max_value=1000)
    )
    def test_positive_odd_integer(self, visitor, base, expo, mono_base, expected):
        mono = self._result_with_base_expo(visitor, base, mono_base, I(None, None), 2*expo+1)
        assert mono == expected

    @pytest.mark.parametrize('mono_base,expected', [
        (M.Nondecreasing, M.Nonincreasing),
        (M.Nonincreasing, M.Nondecreasing),
    ])
    @given(
        base=expressions(),
        expo=st.integers(min_value=1, max_value=1000)
    )
    def test_negative_odd_integer(self, visitor, base, expo, mono_base, expected):
        mono = self._result_with_base_expo(
            visitor, base, mono_base, I(None, None), -2*expo+1
        )
        assert mono == expected

    @pytest.mark.parametrize('mono_base', [M.Nondecreasing, M.Nondecreasing])
    @given(
        base=expressions(),
        expo=reals(allow_infinity=False, min_value=-1e5, max_value=1e5),
    )
    def test_noninteger_negative_base(self, visitor, base, expo, mono_base):
        assume(not almosteq(expo,  0))
        assume(not almosteq(expo, int(expo)))
        mono = self._result_with_base_expo(
            visitor, base, mono_base, I(None, 0), expo
        )
        assert mono == M.Unknown

    @pytest.mark.parametrize('mono_base,expected', [
        (M.Nondecreasing, M.Nondecreasing),
        (M.Nonincreasing, M.Nonincreasing)
    ])
    @given(
        base=expressions(),
        expo=reals(allow_infinity=False, min_value=1e-5, max_value=1e-5)
    )
    def test_positive_noninteger(self, visitor, base, expo, mono_base, expected):
        mono = self._result_with_base_expo(visitor, base, mono_base, I(0, None), expo)
        assert mono == expected

    @pytest.mark.parametrize('mono_base,expected', [
        (M.Nonincreasing, M.Nondecreasing),
        (M.Nondecreasing, M.Nonincreasing)
    ])
    @given(
        base=expressions(),
        expo=reals(allow_infinity=False, min_value=-1e5, max_value=-1e-5)
    )
    def test_negative_noninteger(self, visitor, base, expo, mono_base, expected):
        mono = self._result_with_base_expo(visitor, base, mono_base, I(0, None), expo)
        assert mono == expected


@pytest.mark.parametrize('mono_base,bounds_base,mono_expo,bounds_expo',
    itertools.product(
        [M.Nonincreasing, M.Nondecreasing, M.Unknown],
        [I(None, 0), I(0, None), I(None, None)],
        [M.Nonincreasing, M.Nondecreasing, M.Unknown],
        [I(None, 0), I(0, None), I(None, None)],
    )
)
@given(base=expressions(), expo=expressions())
def test_pow(visitor, base, expo, mono_base, bounds_base, mono_expo, bounds_expo):
    mono = ComponentMap()
    mono[base] = mono_base
    mono[expo] = mono_expo
    bounds = ComponentMap()
    bounds[base] = bounds_base
    bounds[expo] = bounds_expo
    expr = PowExpression([base, expo])
    matched, result = visitor.visit_expression(expr, mono, bounds)
    assert matched
    assert result == M.Unknown


class TestSin(object):
    def _result_with_mono_bounds(self, visitor, g, mono_g, bounds_g):
        mono = ComponentMap()
        mono[g] = mono_g
        bounds = ComponentMap()
        bounds[g] = bounds_g
        expr = pe.sin(g)
        matched, result = visitor.visit_expression(expr, mono, bounds)
        assert matched
        return result

    @given(expressions(), nonnegative_cos_bounds())
    def test_nondecreasing_nonnegative_cos(self, visitor, g, bounds):
        mono = self._result_with_mono_bounds(visitor, g, M.Nondecreasing, bounds)
        assert mono.is_nondecreasing()

    @given(expressions(), nonnegative_cos_bounds())
    def test_nonincreasing_nonnegative_cos(self, visitor, g, bounds):
        mono = self._result_with_mono_bounds(visitor, g, M.Nonincreasing, bounds)
        assert mono.is_nonincreasing()

    @given(expressions(), nonpositive_cos_bounds())
    @pytest.mark.skip("Need investigation")
    def test_decreasing_nonpositive_cos(self, visitor, g, bounds):
        mono = self._result_with_mono_bounds(visitor, g, M.Nondecreasing, bounds)
        assert mono.is_nonincreasing()

    @given(expressions(), nonpositive_cos_bounds())
    def test_nonincreasing_nonpositive_cos(self, visitor, g, bounds):
        mono = self._result_with_mono_bounds(visitor, g, M.Nonincreasing, bounds)
        assert mono.is_nondecreasing()


class TestCos(object):
    def _result_with_mono_bounds(self, visitor, g, mono_g, bounds_g):
        mono = ComponentMap()
        mono[g] = mono_g
        bounds = ComponentMap()
        bounds[g] = bounds_g
        expr = pe.cos(g)
        matched, result = visitor.visit_expression(expr, mono, bounds)
        assert matched
        return result

    @given(expressions(), nonnegative_sin_bounds())
    def test_nonincreasing_nonnegative_sin(self, visitor, g, bounds):
        mono = self._result_with_mono_bounds(visitor, g, M.Nonincreasing, bounds)
        assert mono.is_nondecreasing()

    @given(expressions(), nonnegative_sin_bounds())
    def test_nondecreasing_nonnegative_sin(self, visitor, g, bounds):
        mono = self._result_with_mono_bounds(visitor, g, M.Nondecreasing, bounds)
        assert mono.is_nonincreasing()

    @given(expressions(), nonpositive_sin_bounds())
    def test_nondecreasing_nonpositive_sin(self, visitor, g, bounds):
        mono = self._result_with_mono_bounds(visitor, g, M.Nondecreasing, bounds)
        assert mono.is_nondecreasing()

    @given(expressions(), nonpositive_sin_bounds())
    def test_nonincreasing_nonpositive_sin(self, visitor, g, bounds):
        mono = self._result_with_mono_bounds(visitor, g, M.Nonincreasing, bounds)
        assert mono.is_nonincreasing()



# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SUSPECT math mode.


Selecting Math Mode
-------------------

To set the math mode used by SUSPECT you need to call the ``set_math_mode`` function:


.. code-block:: python

   from suspect.math import set_math_mode, MathMode

   set_math_mode(MathMode.ARBITRARY_PRECISION)


Adding a new Math Mode
----------------------

A math mode needs to define two constants:

 * ``inf``: infinity value
 * ``pi``: the trig constant
 * ``zero``: 0.0

It also needs to define the following functions:

 * ``make_number``: return a new number
 * ``isnan``: predicate to test if it's Not-a-Number
 * ``almosteq``: predicate to test if two numbers are equal
 * ``almostgte```: predicate to test if two numbers are >=
 * ```almostlte``: predicate to test if two numbers are <=
 * ``down``: perform computation rounding down
 * ``up``: perform computation rounding up

The following functions take two parameters, a number and a RoundMode:

 * ``sqrt``
 * ``log``
 * ``exp``
 * ``sin``
 * ``cos``
 * ``tan``
 * ``asin``
 * ``acos``
 * ``atan``
"""

# TODO(fracek): make math mode open like the remaining of suspect.
# TODO(fracek): Then add plugin for correctly rounded arithmetic.


class MathMode:
    """Math mode used internally by SUSPECT."""
    ARBITRARY_PRECISION = 1
    FLOATING_POINT = 2


class RoundMode:
    """Round mode to use for computation.

     * RN: round to nearest
     * RU: round towards +inf
     * RD: round towards -inf
     * RZ: round to zero
    """
    RN = 1
    RU = 2
    RD = 3
    RZ = 4


# pylint: disable=undefined-all-variable
_COMMON_MEMBERS = [
    'make_number',
    'inf',
    'pi',
    'zero',
    'sin',
    'cos',
    'sqrt',
    'log',
    'exp',
    'power',
    'sin',
    'asin',
    'cos',
    'acos',
    'tan',
    'atan',
    'isnan',
    'isinf',
    'almosteq',
    'almostgte',
    'almostlte',
    'down',
    'up',
    'min_',
    'max_',
]

# TODO(fracek): drop this, for now it's need for compat with old code
_ARBITRARY_PRECISION_MEMBERS = ['mpf']

__all__ = _COMMON_MEMBERS


def set_math_mode(math_mode):
    """Set the math mode used by SUSPECT.

    Parameters
    ----------
    math_mode: MathMode
        the math mode to use
    """
    if math_mode == MathMode.ARBITRARY_PRECISION:
        from suspect.math import arbitrary_precision as arb
        for member in _COMMON_MEMBERS:
            globals()[member] = getattr(arb, member)
        for member in _ARBITRARY_PRECISION_MEMBERS:
            globals()[member] = getattr(arb, member)
    elif math_mode == MathMode.FLOATING_POINT:
        from suspect.math import floating_point as fp
        for member in _COMMON_MEMBERS:
            globals()[member] = getattr(fp, member)
    else:
        raise RuntimeError('Invalid MathMode')


set_math_mode(MathMode.FLOATING_POINT)



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# from suspect.monotonicity.monotonicity import Monotonicity
# from suspect.convexity.convexity import Convexity

# This is an hack while we transition to the new visitors. Remove ASAP.
# We want to be able to access values in ctx both as ctx.mono[exp] and
# as ctx.mono(expr)
class _HackDict:
    def __init__(self):
        self._d = {}

    def get(self, key, default=None):
        return self._d.get(key, default)

    def __call__(self, key):
        return self._d[key]

    def __getitem__(self, key):
        return self._d[key]

    def __contains__(self, key):
        return key in self._d

    def __setitem__(self, key, value):
        self._d[key] = value


class SpecialStructurePropagationContext(object):
    def __init__(self):
        self.bounds = _HackDict()
        self.bound = self.bounds # TODO(fracek): for now, remove this later
        self.monotonicity = _HackDict()
        self.convexity = _HackDict()
        self.polynomial = _HackDict()

    def set_bounds(self, expr, value):
        self.bounds[expr] = value

    def get_bounds(self, expr):
        return self.bounds.get(expr)

    def set_monotonicity(self, expr, value):
        self.monotonicity[expr] = value

    def set_convexity(self, expr, value):
        self.convexity[expr] = value



import pytest
import pyomo.environ as pe
from suspect.pyomo import create_connected_model
from suspect.fbbt import perform_fbbt
from suspect.pyomo.quadratic import enable_standard_repn_for_quadratic_expression


enable_standard_repn_for_quadratic_expression()


def model_with_n_variables(n):
    m = pe.ConcreteModel()
    m.I = range(n)
    m.x = pe.Var(m.I, domain=pe.Reals)
    m.y = pe.Var(m.I, domain=pe.Integers, bounds=(1, 10))

    m.obj = pe.Objective(expr=sum(m.x[i]**2.0 for i in m.I))

    num_quad_cons = 10
    @m.Constraint(range(num_quad_cons))
    def quadratic_cons(m, k):
        num_vars = n // num_quad_cons
        return sum(
            m.x[i] * m.x[j]
            for i in range(k*num_vars, (k+1)*num_vars)
            for j in range(k*num_vars, i)
        ) <= 2*n

    @m.Constraint(m.I)
    def bilinear_cons1(m, i):
        return m.x[i]*m.y[i] >= 0

    @m.Constraint(m.I)
    def bilinear_cons2(m, i):
        return m.x[i]*m.y[i] <= m.y[i]

    cm, _ = create_connected_model(m)
    return cm


@pytest.fixture
def small_model():
    return model_with_n_variables(20)


@pytest.fixture
def medium_model():
    return model_with_n_variables(100)


@pytest.fixture
def large_model():
    return model_with_n_variables(300)


def test_initial_fbbt_performance_on_small_model(benchmark, small_model):
    objective_bounds = pe.ComponentMap()
    objective_bounds[small_model.obj.expr] = (None, 100.0)
    result = benchmark(perform_fbbt, small_model, **{'objective_bounds': objective_bounds})


def test_initial_fbbt_performance_on_medium_model(benchmark, medium_model):
    objective_bounds = pe.ComponentMap()
    objective_bounds[medium_model.obj.expr] = (None, 100.0)
    result = benchmark(perform_fbbt, medium_model, **{'objective_bounds': objective_bounds})


@pytest.mark.skip()
def test_initial_fbbt_performance_on_large_model(benchmark, large_model):
    objective_bounds = pe.ComponentMap()
    objective_bounds[large_model.obj.expr] = (None, 100.0)
    result = benchmark(perform_fbbt, large_model, **{'objective_bounds': objective_bounds})


def test_recompute_fbbt_performance_on_small_model(benchmark, small_model):
    objective_bounds = pe.ComponentMap()
    objective_bounds[small_model.obj.expr] = (None, 100.0)
    bounds = perform_fbbt(small_model, objective_bounds=objective_bounds)
    kwargs = {'objective_bounds': objective_bounds, 'initial_bounds': bounds}
    result = benchmark(perform_fbbt, small_model, **kwargs)


def test_recompute_fbbt_performance_on_medium_model(benchmark, medium_model):
    objective_bounds = pe.ComponentMap()
    objective_bounds[medium_model.obj.expr] = (None, 100.0)
    bounds = perform_fbbt(medium_model, objective_bounds=objective_bounds)
    kwargs = {'objective_bounds': objective_bounds, 'initial_bounds': bounds}
    result = benchmark(perform_fbbt, medium_model, **kwargs)


@pytest.mark.skip()
def test_recompute_fbbt_performance_on_large_model(benchmark, large_model):
    objective_bounds = pe.ComponentMap()
    objective_bounds[large_model.obj.expr] = (None, 100.0)
    bounds = perform_fbbt(large_model, objective_bounds=objective_bounds)
    kwargs = {'objective_bounds': objective_bounds, 'initial_bounds': bounds}
    result = benchmark(perform_fbbt, large_model, **kwargs)



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Monotonicity detection rules for nonincreasing functions."""
from suspect.monotonicity.rules.rule import MonotonicityRule


class NonincreasingFunctionRule(MonotonicityRule):
    """Return monotonicity of nonincreasing function."""
    def apply(self, expr, monotonicity, _bounds):
        child = expr.args[0]
        mono = monotonicity[child]
        return mono.negate()


class AcosRule(NonincreasingFunctionRule):
    """Return monotonicity of acos function."""
    pass


class NegationRule(MonotonicityRule):
    """Return monotonicity of negation function."""
    def apply(self, expr, monotonicity, _bounds):
        child = expr.args[0]
        mono = monotonicity[child]
        return mono.negate()



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convexity detector for L2-Norm expressions."""
from suspect.ext import ConvexityDetector
from suspect.convexity import Convexity
from suspect.interfaces import UnaryFunctionRule
from suspect.expression import ExpressionType, UnaryFunctionType


class L2NormConvexityDetector(ConvexityDetector):
    """Detect L_2 Norm convexity in the form

       sqrt(eps + sum_i(x_i**2))
    """
    def __init__(self, _problem):
        super().__init__()

    def visit_expression(self, expr, convexity, mono, bounds):
        return False, None
        return {
            ExpressionType.UnaryFunction: L2NormRule(),
        }


class L2NormRule(UnaryFunctionRule):
    """Detect L_2 Norm expressions."""
    func_type = UnaryFunctionType.Sqrt

    def apply(self, expr, _ctx):
        child = expr.children[0]
        if child.expression_type != ExpressionType.Sum:
            return None

        grandchildren = child.children
        if all([self._square_or_constant(child) for child in grandchildren]):
            return Convexity.Convex
        return None

    def _square_or_constant(self, expr):
        expr_type = expr.expression_type
        if expr_type == ExpressionType.Constant:
            return True
        if expr_type == ExpressionType.Power:
            base, expo = expr.children
            base_is_variable = base.expression_type == ExpressionType.Variable
            expo_is_variable = expo.expression_type == ExpressionType.Constant
            if base_is_variable and expo_is_variable:
                return expo.value == 2
            return False
        if expr_type == ExpressionType.Product:
            if len(expr.children) != 2:
                return False
            f, g = expr.children
            return f is g and f.expression_type == ExpressionType.Variable
        return False



#!/usr/bin/env python

# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import signal
import sys
import time
import logging
import json
import tempfile
import traceback
from mpmath import mpf
import boto3
from suspect.pyomo import create_connected_model, read_osil, enable_standard_repn_for_quadratic_expression
from suspect import (
    detect_special_structure,
    logger,
)

enable_standard_repn_for_quadratic_expression()


class Timeout(object):
    def __init__(self, seconds=1, error_message='Timeout'):
        self.seconds = seconds
        self.error_message = error_message

    def handle_timeout(self, signum, frame):
        raise TimeoutError(self.error_message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)

    def __exit__(self, type, value, traceback):
        signal.alarm(0)


class S3Storage(object):
    PREFIX = 's3://'

    def __init__(self):
        self.s3 = boto3.resource('s3')

    def download_to_temp(self, resource, prefix=None):
        bucket = self._bucket(resource)
        path = self._path(resource)
        temp_dir = tempfile.TemporaryDirectory(prefix=prefix)
        filename = os.path.basename(path)
        dest_file = os.path.join(temp_dir.name, filename)
        bucket.download_file(path, dest_file)
        logging.info('Downloaded {} to {}'.format(resource, dest_file))
        return dest_file, temp_dir

    def upload_file(self, orig, dest):
        bucket = self._bucket(dest)
        path = self._path(dest)
        logging.info('Uploading {} to {}'.format(orig, dest))
        bucket.upload_file(orig, path)

    def _bucket(self, resource):
        resource = resource[len(self.PREFIX):]
        parts = resource.split('/')
        return self.s3.Bucket(parts[0])

    def _path(self, resource):
        resource = resource[len(self.PREFIX):]
        parts = resource.split('/')
        return '/'.join(parts[1:])


class RunResources(object):
    """Context Manager to provide uniform interface for local and S3 objects"""

    def __init__(self, problem, solution, output):
        s3 = S3Storage()
        self.s3 = s3

        if problem.startswith(s3.PREFIX):
            # keep a ref to problem dir so it does not get cleaned self
            self.problem, self.problem_dir = \
                s3.download_to_temp(problem, prefix='problem')
        else:
            self.problem = problem

        if solution is not None and solution.startswith(s3.PREFIX):
            # keep a ref to solution dir so it does not get cleaned self
            self.solution, self.solution_dir = \
                s3.download_to_temp(solution, prefix='solution')
        else:
            self.solution = solution

        self.output_type = None
        if output is None:
            self.output_owned = False
            self.output = sys.stdout
        else:
            self.output_owned = True
            if output.startswith(s3.PREFIX):
                self.output = tempfile.NamedTemporaryFile('w', prefix='output')
                self.output_dest = output
                self.output_type = 's3'
            else:
                self.output = open(output, 'w')
                self.output_type = 'fs'

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if self.output_type == 'fs':
            self.output.close()
        elif self.output_type == 's3':
            self.output.flush()
            self.s3.upload_file(self.output.name, self.output_dest)


def read_problem(filename):
    if filename.endswith('.osil'):
        model = read_osil(
            filename,
            objective_prefix='obj_',
            constraint_prefix='cons_',
        )

        if model.name is None:
            basename = os.path.basename(filename)
            model.name = os.path.splitext(basename)[0]

        return model
    else:
        raise RuntimeError('Unknown file type.')


def _conscurvature(info):
    curvatures = info.conscurvature().values()
    if all(c.is_linear() for c in curvatures):
        return 'linear'
    elif all(c.is_convex() for c in curvatures):
        return 'convex'
    elif all(c.is_concave() for c in curvatures):
        return 'concave'
    else:
        return 'indefinite'


def _objsense(info):
    senses = [o['sense'] for o in info.objectives.values()]
    if len(info.objectives) == 1:
        return senses[0]
    else:
        return ', '.join(senses)


def _objcurvature(info):
    curvatures = [c for c in info.objcurvature().values()]
    if len(info.objectives) == 1:
        cvx = curvatures[0]
        if cvx.is_linear():
            return 'linear'
        elif cvx.is_convex():
            return 'convex'
        elif cvx.is_concave():
            return 'concave'
        else:
            return 'indefinite'
    else:
        return ', '.join(curvatures)


def _objtype(info):
    objtypes = info.objtype().values()
    if len(objtypes) == 1:
        return list(objtypes)[0]
    else:
        return ', '.join(objtypes)


def summarize_problem_structure(info):
    return {
        'name': info.name,
        'nvars': info.num_variables(),
        'nbinvars': info.num_binaries(),
        'nintvars': info.num_integers(),
        'ncons': info.num_constraints(),
        'conscurvature': _conscurvature(info),
        'objsense': _objsense(info),
        'objcurvature': _objcurvature(info),
        'objtype': _objtype(info),
    }


def run_for_problem(filename, solution_filename, timeout_):
    # If we can't read a problem in, don't do anything.
    logging.info('Starting {} / {}. timeout={}'.format(filename, solution_filename, timeout_))
    try:
        with Timeout(seconds=300):
            logging.info('\tReading Problem')
            model = read_problem(filename)
            logging.info('\tConverting DAG')
            model, _ = create_connected_model(model)
    except Exception as err:
        logging.error('{}: Error reading: {}'.format(filename, str(err)))
        return None

    try:
        with Timeout(seconds=timeout_):
            logging.info('\tStarting Special Structure Detection')
            start_t = time.time()
            info = detect_special_structure(model)
            end_t = time.time()
            logging.info('\tSpecial Structure Detection Finished')

    except Exception as err:
        logging.error('{}: {}'.format(model.name, str(err)))
        traceback.print_exc()
        return {
            'name': model.name,
            'status': 'error',
            'error': str(err),
        }

    summary = summarize_problem_structure(info)
    summary['runtime'] = end_t - start_t
    summary['status'] = 'ok'

    if solution_filename is not None:
        try:
            var_bounds_ok, obj_bounds_ok = check_solution_bounds(
                info.variables, info.objectives, solution_filename
            )
            summary['bounds_var_ok'] = var_bounds_ok
            summary['bounds_obj_ok'] = obj_bounds_ok
        except Exception as err:
            logging.error('{}: Solution bounds: {}'.format(model.name, str(err)))

    return summary


def check_solution_bounds(variables, objectives, solution_filename):
    wrong_bounds = []
    objvar_value = None
    with open(solution_filename) as f:
        for line in f:
            parts = line.split()
            var_name = parts[0]
            sol = mpf(parts[1])

            if var_name == 'objvar':
                objvar_value = sol
            if var_name not in variables:
                continue

            v = variables[var_name]

            if sol < v['lower_bound'] or sol > v['upper_bound']:
                wrong_bounds.append(var_name)

    var_bounds_ok = len(wrong_bounds) == 0
    if len(objectives) != 1 or objvar_value is None:
        return var_bounds_ok, None

    for obj in objectives.values():
        lower_bound = obj['lower_bound']
        upper_bound = obj['upper_bound']
        obj_bounds_ok = (
                objvar_value >= lower_bound and objvar_value <= upper_bound
        )
        return var_bounds_ok, obj_bounds_ok


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)

    parser = argparse.ArgumentParser()
    parser.add_argument('--problem', '-p')
    parser.add_argument('--solution', '-s')
    parser.add_argument('--output', '-o', nargs='?')
    parser.add_argument('--timeout', type=int, default=300,
                        help='Timeout in seconds')
    parser.add_argument('--log', dest='log_level',
                        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
                        default='WARNING')
    args = parser.parse_args()

    logger.setLevel(level=args.log_level)

    if args.problem is None:
        parser.print_help()
        sys.exit(1)

    with RunResources(args.problem, args.solution, args.output) as r:
        result = run_for_problem(r.problem, r.solution, args.timeout)

        if result is None:
            sys.exit(1)

        result_str = json.dumps(result, sort_keys=True)
        r.output.write(result_str + '\n')



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Monotonicity rules."""
from .base import VariableRule, ConstantRule, ConstraintRule, ObjectiveRule
from .abs import AbsRule
from .division import DivisionRule, ReciprocalRule
from .linear import LinearRule, SumRule
from .nondecreasing import SqrtRule, ExpRule, LogRule, TanRule, AsinRule, AtanRule
from .nonincreasing import AcosRule, NegationRule
from .pow import PowerRule
from .product import ProductRule
from .quadratic import QuadraticRule
from .trig import SinRule, CosRule


__all__ = [
    'VariableRule', 'ConstantRule', 'ConstraintRule', 'ObjectiveRule',
    'ProductRule', 'DivisionRule', 'LinearRule', 'SumRule', 'AbsRule',
    'SqrtRule', 'ExpRule', 'LogRule', 'TanRule', 'AsinRule', 'AtanRule',
    'AcosRule', 'NegationRule', 'PowerRule', 'SinRule', 'CosRule',
    'QuadraticRule', 'ReciprocalRule'
]



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convexity detection rules for power expressions."""
from suspect.convexity.convexity import Convexity
from suspect.convexity.rules.rule import ConvexityRule
from suspect.pyomo.expressions import nonpyomo_leaf_types
from suspect.math import almosteq, almostgte # pylint: disable=no-name-in-module


class PowerRule(ConvexityRule):
    """Return monotonicity of power expression."""
    def apply(self, expr, convexity, monotonicity, bounds):
        assert len(expr.args) == 2
        base, expo = expr.args

        mono_base = monotonicity[base]
        mono_expo = monotonicity[expo]

        if mono_expo.is_constant():
            if type(expo) not in nonpyomo_leaf_types:
                expo = expo.value
            cvx_base = convexity[base]
            bounds_base = bounds[base]
            return _constant_expo_pow_convexity(expo, cvx_base, bounds_base)
        elif mono_base.is_constant():
            if type(base) not in nonpyomo_leaf_types:
                base = base.value
            cvx_expo = convexity[expo]
            return _constant_base_pow_convexity(base, cvx_expo)
        return Convexity.Unknown


def _constant_base_pow_convexity(base, cvx_expo):
    if 0 < base < 1:
        if cvx_expo.is_concave():
            return Convexity.Convex
    elif almostgte(base, 1):
        if cvx_expo.is_convex():
            return Convexity.Convex
    return Convexity.Unknown


def _constant_expo_pow_convexity(expo, cvx_base, bounds_base):
    is_integer = almosteq(expo, int(expo))
    is_even = almosteq(expo % 2, 0)

    if almosteq(expo, 0):
        return Convexity.Linear
    elif almosteq(expo, 1):
        return cvx_base
    elif is_integer and is_even:
        return _integer_even_expo_pow_convexity(expo, cvx_base, bounds_base)
    elif is_integer:
        return _integer_odd_expo_pow_convexity(expo, cvx_base, bounds_base)
    return _noninteger_expo_pow_convexity(expo, cvx_base, bounds_base)


def _integer_even_expo_pow_convexity(expo, cvx_base, bounds_base):
    if expo > 0:
        return _integer_positive_even_expo_pow_convexity(expo, cvx_base, bounds_base)
    return _integer_negative_even_expo_pow_convexity(expo, cvx_base, bounds_base)


def _integer_positive_even_expo_pow_convexity(_expo, cvx_base, bounds_base):
    if cvx_base.is_linear():
        return Convexity.Convex
    elif cvx_base.is_convex() and bounds_base.is_nonnegative():
        return Convexity.Convex
    elif cvx_base.is_concave() and bounds_base.is_nonpositive():
        return Convexity.Convex
    return Convexity.Unknown


def _integer_negative_even_expo_pow_convexity(_expo, cvx_base, bounds_base):
    if cvx_base.is_convex() and bounds_base.is_nonpositive():
        return Convexity.Convex
    elif cvx_base.is_concave() and bounds_base.is_nonnegative():
        return Convexity.Convex
    elif cvx_base.is_convex() and bounds_base.is_nonnegative():
        return Convexity.Concave
    elif cvx_base.is_concave() and bounds_base.is_nonpositive():
        return Convexity.Concave
    return Convexity.Unknown


def _integer_odd_expo_pow_convexity(expo, cvx_base, bounds_base):
    if expo > 0:
        return _integer_positive_odd_expo_pow_convexity(expo, cvx_base, bounds_base)
    return _integer_negative_odd_expo_pow_convexity(expo, cvx_base, bounds_base)


def _integer_positive_odd_expo_pow_convexity(expo, cvx_base, bounds_base):
    if almosteq(expo, 1): # pragma: no cover
        return cvx_base # we won't reach this because of the check in apply.
    elif cvx_base.is_convex() and bounds_base.is_nonnegative():
        return Convexity.Convex
    elif cvx_base.is_concave() and bounds_base.is_nonpositive():
        return Convexity.Concave
    return Convexity.Unknown


def _integer_negative_odd_expo_pow_convexity(_expo, cvx_base, bounds_base):
    if cvx_base.is_concave() and bounds_base.is_nonnegative():
        return Convexity.Convex
    elif cvx_base.is_convex() and bounds_base.is_nonpositive():
        return Convexity.Concave
    return Convexity.Unknown


def _noninteger_expo_pow_convexity(expo, cvx_base, bounds_base):
    if not bounds_base.is_nonnegative():
        return Convexity.Unknown
    if cvx_base.is_convex() and expo > 1:
        return Convexity.Convex
    elif cvx_base.is_concave() and expo < 0:
        return Convexity.Convex
    elif cvx_base.is_concave() and 0 < expo < 1:
        return Convexity.Concave
    elif cvx_base.is_convex() and expo < 0:
        return Convexity.Concave
    return Convexity.Unknown



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FBBT bounds propagation visitor."""

from suspect.pyomo.expressions import (
    nonpyomo_leaf_types,
    Var,
    NumericConstant,
    Constraint,
    Objective,
    MonomialTermExpression,
    ProductExpression,
    DivisionExpression,
    ReciprocalExpression,
    LinearExpression,
    SumExpression,
    PowExpression,
    NegationExpression,
    AbsExpression,
    UnaryFunctionExpression,
)
from suspect.visitor import ForwardVisitor
from suspect.pyomo.quadratic import QuadraticExpression
from suspect.fbbt.propagation.rules import (
    VariableRule,
    ConstantRule,
    ConstraintRule,
    ObjectiveRule,
    ProductRule,
    DivisionRule,
    ReciprocalRule,
    LinearRule,
    SumRule,
    PowerRule,
    NegationRule,
    AbsRule,
    QuadraticRule,
    UnaryFunctionRule,
)


_expr_to_rule_map = dict()
_expr_to_rule_map[NumericConstant] = ConstantRule()
_expr_to_rule_map[Var] = VariableRule()
_expr_to_rule_map[Constraint] = ConstraintRule()
_expr_to_rule_map[Objective] = ObjectiveRule()
_expr_to_rule_map[ProductExpression] = ProductRule()
_expr_to_rule_map[MonomialTermExpression] = ProductRule()
_expr_to_rule_map[DivisionExpression] = DivisionRule()
_expr_to_rule_map[ReciprocalExpression] = ReciprocalRule()
_expr_to_rule_map[LinearExpression] = LinearRule()
_expr_to_rule_map[SumExpression] = SumRule()
_expr_to_rule_map[PowExpression] = PowerRule()
_expr_to_rule_map[NegationExpression] = NegationRule()
_expr_to_rule_map[AbsExpression] = AbsRule()
_expr_to_rule_map[UnaryFunctionExpression] = UnaryFunctionRule()
_expr_to_rule_map[QuadraticExpression] = QuadraticRule()


def propagate_bounds_leaf_to_root(expr, bounds):
    if type(expr) in nonpyomo_leaf_types:
        rule = _expr_to_rule_map[NumericConstant]
    elif expr.is_constant():
        rule = _expr_to_rule_map[NumericConstant]
    elif expr.is_variable_type():
        rule = _expr_to_rule_map[Var]
    else:
        assert expr.is_expression_type()
        rule = _expr_to_rule_map[type(expr)]
    return rule.apply(expr, bounds)


class BoundsPropagationVisitor(ForwardVisitor):
    """Tighten bounds from sources to sinks."""
    intersect_abs_eps = 1e-10

    def visit_expression(self, expr, bounds):
        return True, propagate_bounds_leaf_to_root(expr, bounds)

    def handle_result(self, expr, new_bounds, bounds):
        old_bounds = bounds.get(expr, None)
        if old_bounds is not None:
            # bounds exists, tighten it
            new_bounds = old_bounds.intersect(
                new_bounds,
                abs_eps=self.intersect_abs_eps,
            )
            has_changed = new_bounds != old_bounds
        else:
            has_changed = True

        bounds[expr] = new_bounds
        return has_changed



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convexity detection rules for abs function."""
from suspect.convexity.convexity import Convexity
from suspect.convexity.rules.rule import ConvexityRule


class AbsRule(ConvexityRule):
    """Return convexity of abs."""
    def apply(self, expr, convexity, _mono, bounds):
        child = expr.args[0]
        cvx = convexity[child]
        bounds = bounds[child]
        if cvx.is_linear():
            return Convexity.Convex
        elif cvx.is_convex():
            if bounds.is_nonnegative():
                return Convexity.Convex
            elif bounds.is_nonpositive():
                return Convexity.Concave
        elif cvx.is_concave():
            if bounds.is_nonpositive():
                return Convexity.Convex
            elif bounds.is_nonnegative():
                return Convexity.Concave
        return Convexity.Unknown
